/*Copyright (C) 2009  Annette Bieniusa, Arie Middelkoop.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package vacation;

import java.util.EnumMap;
import java.util.Random;

import dstm.Atomic;

import vacation.Reservation.ReservationType;

public class Client extends Thread {
	
	public enum Action { MAKE_RESERVATION, DELETE_CUSTOMER, UPDATE_TABLES};
	
	int id;
	Manager manager;
	Random randomgen;
	int numOperation;
	int numQueryPerTransaction;
	int queryRange;
	int percentUser;
	
	public Client(int id, Manager manager,int numOperation, int numQueryPerTransaction, int queryRange, int percentUser){
		this.id = id;
		this.manager = manager;
		this.randomgen = new Random();
		this.numOperation = numOperation;
		this.numQueryPerTransaction = numQueryPerTransaction;
		this.queryRange = queryRange;
		this.percentUser = percentUser;	
	}
	
	/**
	 * Selects and action according to the specified distribution.
	 * @param r random number
	 * @param percentUser percentage of reservations
	 * @return
	 */
	private Action selectAction (int r, int percentUser){
		if (r < percentUser) {
			return Action.MAKE_RESERVATION;
		} else if (Integer.lowestOneBit(r) == 0) {
			return Action.DELETE_CUSTOMER;
		} else {
			return Action.UPDATE_TABLES;
		}
	}
	
	public void run(){
		ReservationType[] types  = new ReservationType[this.numQueryPerTransaction];
		int[] ids = new int[this.numQueryPerTransaction];
		boolean[] ops = new boolean[this.numQueryPerTransaction];
		int[] prices = new int[this.numQueryPerTransaction];
		
		for (int i = 0; i < this.numOperation; i++){
			int r = this.randomgen.nextInt(100);
			Action action = this.selectAction(r, percentUser);
			int customerId = this.randomgen.nextInt(queryRange) + 1;
			
			
			switch(action) {
			case MAKE_RESERVATION : 
				EnumMap<ReservationType,Integer> maxPrices = new EnumMap<ReservationType,Integer>(ReservationType.class);
				for (ReservationType t : ReservationType.values()){
					maxPrices.put(t, -1);
				}
				EnumMap<ReservationType,Integer> maxIds = new EnumMap<ReservationType,Integer>(maxPrices);
				
				int numQuery = this.randomgen.nextInt(numQueryPerTransaction) + 1;
				
				for (int n = 0; n < numQuery; n++) {
					types[n] = ReservationType.values()[this.randomgen.nextInt(ReservationType.values().length)];
					ids[n] = this.randomgen.nextInt(queryRange) + 1;
				}
				makeReservations(types, ids, customerId, maxPrices, maxIds,
							numQuery);
				
				break;
			case DELETE_CUSTOMER  :
				deleteCustomer(customerId);
				break;
			case UPDATE_TABLES    :
				int numUpdate = this.randomgen.nextInt(this.numQueryPerTransaction) + 1;
				for (int n = 0; n < numUpdate; n++){
					types[n] = ReservationType.values()[this.randomgen.nextInt(ReservationType.values().length)];
					ids[n] = this.randomgen.nextInt(this.queryRange) + 1;
					ops[n] = this.randomgen.nextBoolean();
					if (ops[n]) {
						prices[n] = (this.randomgen.nextInt(5) * 10) + 50;
					}
				}
				updateTables(types, ids, ops, prices, numUpdate);
				break;
			}
		}
		//Thread ends here
	}

	@Atomic
	private void updateTables(ReservationType[] types, int[] ids,
			boolean[] ops, int[] prices, int numUpdate) {
		for (int n = 0; n < numUpdate; n++) {
			ReservationType t = types[n];
			int id = ids[n];
			boolean doAdd = ops[n];
			if (doAdd) {
				int newPrice = prices[n];
				switch(t){
				case CAR:    this.manager.addCar(id, 100, newPrice); break;
				case FLIGHT: this.manager.addFlight(id, 100, newPrice); break;
				case ROOM:   this.manager.addRoom(id, 100, newPrice); break;
				// add Exception
				}
			}
			else {
				switch(t) {
				case CAR:    this.manager.deleteCar(id, 100); break;
				case FLIGHT: this.manager.deleteFlight(id); break;
				case ROOM:   this.manager.deleteRoom(id, 100); break;
				// add Exception
				}	
			}
			//transaction ends			
		}
	}


	@Atomic
	private void deleteCustomer(int customerId) {
		int bill = this.manager.queryCustomerBill(customerId);
		if (bill >= 0) {
			this.manager.deleteCustomer(customerId);
		}
	}


	@Atomic
	private void makeReservations(ReservationType[] types, int[] ids,
			int customerId, EnumMap<ReservationType, Integer> maxPrices,
			EnumMap<ReservationType, Integer> maxIds, int numQuery) {
		boolean isFound = false;
		for (int hugo = 0; hugo < numQuery; ++hugo) {
			ReservationType t = types[hugo];
			int id = ids[hugo];
			int price = -1;
			switch(t){
			case CAR :
				if (this.manager.queryCar(id) >= 0) {					
					price = this.manager.queryCarPrice(id);
				}
				break;
			case FLIGHT :
				if (this.manager.queryFlight(id) >= 0) {
					price = this.manager.queryFlightPrice(id);
				}
				break;
			case ROOM :				
				if (this.manager.queryRoom(id) >= 0) {
					price = this.manager.queryRoomPrice(id);
				}
				break;	
			//add a good exception
			}
			
			if (price > maxPrices.get(t)) {	
				maxPrices.put(t, price);
				maxIds.put(t, id);
				isFound = true;
			}
		}
		
		if (isFound) {
			this.manager.addCustomer(customerId);
		}
		if (maxIds.get(ReservationType.CAR) > 0) {
			this.manager.reserveCar(customerId, id);
		}
		if (maxIds.get(ReservationType.FLIGHT) > 0) {
			this.manager.reserveFlight(customerId, id);
		}
		if (maxIds.get(ReservationType.ROOM) > 0) {
			this.manager.reserveRoom(customerId, id);
		}
	}
}