/*Copyright (C) 2009  Annette Bieniusa, Arie Middelkoop.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package vacation;

import dstm.Atomic;
import vacation.Reservation.ReservationType;

public class Manager {

	RBTree<Integer, Reservation> carTable;
	RBTree<Integer, Reservation> roomTable;
	RBTree<Integer, Reservation> flightTable;
	RBTree<Integer, Customer> customerTable;

	// This is the substitute for the constructor, as we cannot initialize these
	// transactional data structures in a constructor
	@Atomic
	public static Manager init() {
		Manager m = new Manager();
		m.carTable = new RBTree<Integer, Reservation>();
		m.roomTable = new RBTree<Integer, Reservation>();
		m.flightTable = new RBTree<Integer, Reservation>();
		m.customerTable = new RBTree<Integer, Customer>();
		return m;
	}

	public enum Task {
		CAR, FLIGHT, ROOM, CUSTOMER;
	};

	public RBTree<Integer, Reservation> getAssociatedReservationTable(
			ReservationType t) {
		RBTree<Integer, Reservation> table = null;
		switch (t) {
		case CAR:
			table = (RBTree<Integer, Reservation>) carTable;
			break;
		case FLIGHT:
			table = (RBTree<Integer, Reservation>) flightTable;
			break;
		case ROOM:
			table = (RBTree<Integer, Reservation>) roomTable;
			break;
		}
		return table;
	}

	public boolean addTask(Task t, int id, int num, int price) {
		boolean b = false;
		switch (t) {
		case CAR:
			b = this.addCar(id, num, price);
			break;
		case FLIGHT:
			b = this.addFlight(id, num, price);
			break;
		case ROOM:
			b = this.addRoom(id, num, price);
			break;
		case CUSTOMER:
			b = this.addCustomer(id);
			break;
		}
		return b;
	}

	@Atomic
	private boolean addReservation(RBTree<Integer, Reservation> table, int id,
			int num, int price) {
		Reservation reservation = table.get(id);
		if (reservation == null) {
			if (num < 1 || price < 0) {
				return false;
			}
			reservation = new Reservation(id, num, price);
			table.put(id, reservation);
		} else {
			if (!reservation.addToTotal(num)) {
				return false;
			}
			if (reservation.getNumTotal() == 0) {
				table.remove(id);
			} else {
				reservation.updatePrice(price);
			}
		}
		return true;
	}

	public boolean addCar(int id, int num, int price) {
		return this.addReservation(this.carTable, id, num, price);
	}

	public boolean deleteCar(int id, int num) {
		// -1 keeps old price
		return this.addReservation(this.carTable, id, -num, -1);
	}

	public boolean addRoom(int id, int num, int price) {
		return this.addReservation(this.roomTable, id, num, price);

	}

	public boolean deleteRoom(int id, int num) {
		return this.addReservation(this.roomTable, id, -num, -1);
	}

	public boolean addFlight(int id, int num, int price) {
		return this.addReservation(this.flightTable, id, num, price);
	}

	@Atomic
	public boolean deleteFlight(int id) {
		Reservation reservation = flightTable.get(id);
		if (reservation == null) {
			return false;
		}
		synchronized (reservation) {
			if (reservation.getNumUsed() > 0) {
				return false;
			}
			return (this.addReservation(this.flightTable, id, -1
					* reservation.getNumTotal(), -1));
		}
	}

	@Atomic
	public boolean addCustomer(int id) {
		Customer customer = this.customerTable.get(id);
		if (customer != null) {
			return false;
		}

		SingleLinkedList<ReservationInfo> resInfoList = new SingleLinkedList<ReservationInfo>();
		resInfoList.init();
		customer = new Customer(id, resInfoList);
		this.customerTable.put(id, customer);
		return true;
	}

	@Atomic
	public boolean deleteCustomer(int id) {
		Customer customer = null;
		customer = this.customerTable.get(id);

		if (customer == null) {
			return false;
		}

			
		for (ReservationInfo resInfo : customer.reservationInfoList){
			RBTree<Integer, Reservation> table = this
					.getAssociatedReservationTable(resInfo.getType());
			Reservation reservation = table.get(resInfo.getId());
			reservation.cancel();
		}
		this.customerTable.remove(id);
		return true;
	}

	@Atomic
	private int queryNumFree(RBTree<Integer, Reservation> table, int id) {
		int numFree = -1;
		Reservation reservation = table.get(id);
		if (reservation != null) {
			numFree = reservation.getNumFree();
		}
		return numFree;
	}

	@Atomic
	private int queryPrice(RBTree<Integer, Reservation> table, int id) {
		int price = -1;
		Reservation reservation = table.get(id);
		if (reservation != null) {
			price = reservation.getPrice();
		}
		return price;
	}

	public int queryCar(int id) {
		return this.queryNumFree(carTable, id);
	}

	public int queryCarPrice(int id) {
		return this.queryPrice(carTable, id);
	}

	public int queryRoom(int id) {
		return this.queryNumFree(roomTable, id);
	}

	public int queryRoomPrice(int id) {
		return this.queryPrice(roomTable, id);
	}

	public int queryFlight(int id) {
		return this.queryNumFree(flightTable, id);
	}

	public int queryFlightPrice(int id) {
		return this.queryPrice(flightTable, id);
	}

	@Atomic
	public int queryCustomerBill(int id) {
		int bill = -1;
		Customer customer = this.customerTable.get(id);
		if (customer != null) {
			bill = customer.getBill();
		}
		return bill;
	}

	@Atomic
	private boolean reserve(RBTree<Integer, Reservation> table, int customerId,
			int id, ReservationType t) {
		Customer customer = this.customerTable.get(customerId);
		if (customer == null) {
			return false;
		}

		Reservation reservation = table.get(id);
		if (reservation == null) {
			return false;
		}

		if (!reservation.make()) {
			return false;
		}

		/* Undo previous successful reservation */
		if (!customer.addReservationInfo(t, id, reservation.getPrice())) {
			reservation.cancel();
			return false;
		}
		return true;
	}

	public boolean reserveCar(int customerId, int carId) {
		return this.reserve(this.carTable, customerId, carId,
				ReservationType.CAR);
	}

	public boolean reserveRoom(int customerId, int roomId) {
		return this.reserve(this.roomTable, customerId, roomId,
				ReservationType.ROOM);
	}

	public boolean reserveFlight(int customerId, int flightId) {
		return this.reserve(this.flightTable, customerId, flightId,
				ReservationType.FLIGHT);
	}

	@Atomic
	private boolean cancel(RBTree<Integer, Reservation> table, int customerId,
			int id, ReservationType t) {
		Customer customer = this.customerTable.get(customerId);
		if (customer == null) {
			return false;
		}

		Reservation reservation = table.get(id);
		if (reservation == null) {
			return false;
		}

		/* Undo previous successful cancellation */
		if (!customer.removeReservationInfo(t, id)) {
			reservation.make();
			return false;
		}
		return true;
	}

	public boolean cancelCar(int customerId, int carId) {
		return this.cancel(this.carTable, customerId, carId,
				ReservationType.CAR);
	}

	public boolean cancelRoom(int customerId, int roomId) {
		return this.cancel(this.roomTable, customerId, roomId,
				ReservationType.ROOM);
	}

	public boolean cancelFlight(int customerId, int flightId) {
		return this.cancel(this.flightTable, customerId, flightId,
				ReservationType.FLIGHT);
	}

	public RBTree<Integer, Customer> getCustomerTable() {
		return this.customerTable;
	}

}
