/*Copyright (C) 2009  Annette Bieniusa, Arie Middelkoop.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package vacation;

import dstm.Atomic;
import dstm.Transactional;

@Transactional
public class RBTree<K extends Comparable<K>,V> {
	
	private static enum Color {
		Red, Black;
	}
	
	@Transactional
	private class Node {
		K key;
		V value;
		Node parent, left, right;
		Color c;
		
		@Atomic
		public String toString() {
			String s ="";
			if (left != null) {
				s += left.toString() + "  ";
			}
			if (parent != null) {
				s += "(" + key +","+ value +"," + c + "," + parent.key + ")";
			} else {
				s += "(" + key +"," + value +","+ c + ",  )";	
			}
			if (right != null) {
				s += "  " + right.toString();
			}
			return s;
		}
		
		@Atomic
		public int size() {
			int s = 1;
			if (left != null) {
				s += left.size();
			}
			if (right != null) {
				s += right.size();
			}
			return s;
		}
	}
	
	private Node root;
	
	@Atomic
	private Node lookup (K k) {
		Node current = this.root;
		
		while (current != null) {
			int cmp = k.compareTo(current.key);
			if (cmp == 0) {
				return current;
			}
			current = (cmp < 0) ? current.left : current.right;
		}
		return null;
	}
	
	@Atomic
	private void rotateLeft (Node x) {
		Node r = x.right;
		Node rl = r.left;
		
		x.right = rl;
		if (rl != null) {
			rl.parent = x;
		}
		
		Node xp = x.parent;
		r.parent = xp;
		if (xp == null) {
			this.root = r;
		} else if (xp.left == x) {
			xp.left = r;
		} else {
			xp.right = r;
		}
		r.left = x;
		x.parent = r;
	}
	
	@Atomic
	private void rotateRight (Node x) {
		Node l = x.left;
		Node lr = l.right;
		
		x.left = lr;
		if (lr != null) {
			lr.parent = x;
		}
		
		Node xp = x.parent;
		l.parent = xp;
		if (xp == null) {
			this.root = l;
		} else if (xp.right == x) {
			xp.right = l;
		} else {
			xp.left = l;
		}
		l.right = x;
		x.parent = l;
	}
	
	@Atomic
	private Node parentOf (Node n) {
		return (n == null)? null : n.parent;
	}
	
	@Atomic
	private Node leftOf (Node n) {
		return (n == null)? null : n.left;
	}

	@Atomic
	private Node rightOf (Node n) {
		return (n == null)? null : n.right;
	}
	
	@Atomic
	private Color colorOf (Node n) {
		return (n == null)? Color.Black : n.c;
	}
	
	@Atomic
	private void setColor(Node n, Color c) {
		if (n != null) {
			n.c = c;
		}
	}
	
	@Atomic
	private void fixAfterInsertion (Node x) {
		x.c = Color.Red;
		while (x != null && x != this.root) {
			Node xp = x.parent;
			if (xp.c != Color.Red) {
				break;
			}
			if (parentOf(x) == leftOf(parentOf(parentOf(x)))) {
				Node y = rightOf(parentOf(parentOf(x)));
				if (colorOf(y) == Color.Red) {
					setColor(parentOf(x), Color.Black);
					setColor(y,Color.Black);
					setColor(parentOf(parentOf(x)), Color.Red);
					x = parentOf(parentOf(x));
				} else {
					if (x == rightOf(parentOf(x))) {
						x = parentOf(x);
						rotateLeft(x);
					}
					setColor(parentOf(x), Color.Black);
					setColor(parentOf(parentOf(x)), Color.Red);
					if (parentOf(parentOf(x)) != null ) {
						rotateRight(parentOf(parentOf(x)));
					}
				}
			} else {
				Node y = leftOf(parentOf(parentOf(x)));
				if (colorOf(y) == Color.Red) {
					setColor(parentOf(x), Color.Black);
					setColor(y, Color.Black);
					setColor(parentOf(parentOf(x)),Color.Red);
					x = parentOf(parentOf(x));
				} else {
					if (x == leftOf(parentOf(x))) {
						x = parentOf(x);
						rotateRight(x);
					}
					setColor(parentOf(x), Color.Black);
					setColor(parentOf(parentOf(x)), Color.Red);
					if (parentOf(parentOf(x)) != null ) {
						rotateLeft(parentOf(parentOf(x)));
					}
				}
			}
		}
		Node ro = this.root;
		if (ro.c != Color.Black) {
			ro.c = Color.Black;
		}
	}
	
	@Atomic
	private Node insert (K k, V v, Node n) {
		Node t = this.root;
		if (t == null) {
			if (n == null) {
				return null;
			}	
			n.left = null;
			n.right = null;
			n.parent = null;
			n.key = k;
			n.value = v;
			n.c = Color.Black;
			this.root = n;
			return null;
		}
		while (true) {
			int cmp = k.compareTo(t.key);
			if (cmp == 0) {
				return t;
			} else if (cmp < 0) {
				Node tl = t.left;
				if (tl != null) {
					t = tl;
				} else {
					n.left = null;
					n.right = null;
					n.key = k;
					n.value = v;
					n.parent = t;
					t.left = n;
					fixAfterInsertion(n);
					return null;
				}
			} else {
				Node tr = t.right;
				if (tr != null) {
					t = tr;
				} else {
					n.left = null;
					n.right = null;
					n.key = k;
					n.value = v;
					n.parent = t;
					t.right = n;
					fixAfterInsertion(n);
					return null;
				}
			}
		}
	}
	
	@Atomic
	private Node successor (Node t) {
		if (t == null) {
			return null;
		} else if (t.right != null) {
			Node current = t.right;
			while (current.left != null) {
				current = current.left;
			}
			return current;
		} else {
			Node p = t.parent;
			Node ch = t;
			while (p != null && ch == p.right) {
				ch = p;
				p = p.parent;
			}
			return p;
		}
	}
	
	@Atomic
	private void fixAfterDeletion (Node x) {
		while (x != this.root && colorOf(x) == Color.Black) {
			if (x == leftOf(parentOf(x))) {
				Node sib = rightOf(parentOf(x));
				if (colorOf(sib) == Color.Red) {
					setColor(sib,Color.Black);
					setColor(parentOf(x), Color.Red);
					rotateLeft(parentOf(x));
					sib = rightOf(parentOf(x));
				}
				if (colorOf(leftOf(sib)) == Color.Black && colorOf(rightOf(sib)) == Color.Black) {
					setColor(sib, Color.Red);
					x = parentOf(x);
				} else {
					if (colorOf(rightOf(sib)) == Color.Black) {
						setColor(leftOf(sib), Color.Black);
						setColor(sib,Color.Red);
						rotateRight(sib);
						sib = rightOf(parentOf(x));
					}
					setColor(sib, colorOf(parentOf(x)));
					setColor(parentOf(x), Color.Black);
					setColor(rightOf(sib), Color.Black);
					rotateLeft(parentOf(x));
					x = this.root;
				}
			} else {
				Node sib = leftOf(parentOf(x));
				if (colorOf(sib) == Color.Red) {
					setColor(sib, Color.Black);
					setColor(parentOf(x), Color.Red);
					rotateRight(parentOf(x));
					sib = leftOf(parentOf(x));
				}
				if (colorOf(rightOf(sib)) == Color.Black && colorOf(leftOf(sib)) == Color.Black) {
					setColor(sib, Color.Red);
					x = parentOf(x);
				} else {
					if ( colorOf(leftOf(sib)) == Color.Black) {
						setColor(rightOf(sib), Color.Black);
						setColor(sib,Color.Red);
						rotateLeft(sib);
						sib = leftOf(parentOf(x));
					}
					setColor(sib, colorOf(parentOf(x)));
					setColor(parentOf(x), Color.Black);
					setColor(leftOf(sib), Color.Black);
					rotateRight(parentOf(x));
					x = this.root;
				}
			}
		}
		if (x != null && x.c != Color.Black) {
			x.c = Color.Black;
		}
	}
	
	@Atomic
	private Node delete (Node p) {
		if (p.left != null && p.right != null) {
			Node s = successor(p);
			p.key = s.key;
			p.value = s.value;
			p = s;
		}
		
		Node replacement = (p.left != null) ? p.left : p.right;
		if (replacement != null) {
			replacement.parent = p.parent;
			Node pp = p.parent;
			if (pp == null) {
				this.root = replacement;
			} else if (p == pp.left) {
				pp.left = replacement;
			} else {
				pp.right = replacement;
			}
			
			p.left = null;
			p.right = null;
			p.parent = null;
			
			if (p.c == Color.Black) {
				fixAfterDeletion(replacement);
			}
		} else if(p.parent == null) {
			this.root = null;
		} else {
			if (p.c == Color.Black) {
				fixAfterDeletion(p);
			}
			Node pp = p.parent;
			if (pp != null) {
				if (p == pp.left) {
					pp.left = null;
				} else if (p == pp.right) {
					pp.right = null;
				}
				p.parent = null;
			}
		}
		return p;
	}
	
	/**
	 * Adds a node with the key and value.
	 * @param key
	 * @param value
	 * @return true if node was added successfully, false if node was not added because key already existed
	 */
	@Atomic
	public boolean put (K key, V value) {
		Node n = new Node();
		Node ex = insert(key,value,n);
		return (ex == null ? true : false);
	}
	
	/**
	 * Deletes the node containing the key.
	 * @param key
	 * @return true if node was deleted, false if the key was not in tree
	 */
	@Atomic
	public boolean remove (K key) {
		Node n = lookup(key);
		if (n != null) {
			n = delete(n);
		}
		return (n != null ? true : false);
	}
	
	
	/**
	 * Updates the value associated to a key.
	 * @param key
	 * @param value
	 * @return true if the key was already in the tree, false if a new node got inserted
	 */
	@Atomic
	public boolean update (K key, V value) {
		Node nn = new Node();
		Node ex = insert(key,value, nn);
		if (ex != null) {
			ex.value = value;
			return true;
		}
		return false;
	} 
	
	/**
	 * Returns the value that is associated with the key.
	 * @param key
	 * @return value associated with the key, otherwise null
	 */
	@Atomic
	public V get (K key) {
		Node n = lookup(key);
		if (n != null) {
			return n.value;
		}
		return null;
	}
	/**
	 * Checks if a key is in the tree.
	 * @param key
	 * @return true if the key is in the tree, otherwise false
	 */
	@Atomic 
	public boolean find (K key) {
		Node n = lookup(key);
		return (n != null);
	}
	
	
	/**
	 * Returns a in-order string representation of the tree.
	 * A node is represented as a quadruple of key, value, color and key of the parent node.
	 */
	@Atomic
	public String toString() {
		if (root == null) {
			return "";
		} else
			return root.toString();
	}
	
	/**
	 * Gives the size of the tree.
	 * @return number of nodes
	 */
	@Atomic
	public int size () {
		if (root == null) {
			return 0;
		} else
			return root.size();
	}
	
	
	/*
	// Testing of sequential code
	private Node firstEntry() {
		Node p = this.root;
		if (p != null) {
			while (p.left != null) {
				p = p.left;
			}
		}
		return p;
	}
	
	private int verifyRedBlack (Node n, int depth) {
		if (n == null) {
			return 1;
		}
		
		int heightLeft = verifyRedBlack(n.left, depth+1);
		int heightRight = verifyRedBlack(n.right, depth+1);
		
		if (heightLeft == 0 || heightRight == 0) {
			return 0;
		}
		
		if (heightLeft != heightRight) {
			System.out.println("Imbalance at depth=" + depth +": heightLeft="+heightLeft+" and heightRight=" + heightRight);
		}
		
		if (n.left != null && n.left.parent != n) {
			System.out.println("Lineage is broken\n");
		}
		
		if (n.right != null && n.right.parent != n) {
			System.out.println("Lineage is broken\n");
		}
		
		if (n.c == Color.Red) {
			if (n.left != null && n.left.c != Color.Black) {
				System.out.println("Verify?!");
				return 0;
			}
			if (n.right != null && n.right.c != Color.Black) {
				System.out.println("Verify?!");
				return 0;
			}
			return heightLeft;
		}
		if (n.c != Color.Black) {
			System.out.println("Verify?!");
			return 0;
		}
		return (heightLeft + 1);
	}
	
	private int verify () {
		if (this.root == null) {
			return 1;
		}
		
		//System.out.println("Integrity check: ");
		
		if (this.root.parent != null) {
			System.out.println(" root has parent " + root.parent);
			return -1;
		}
		
		if (this.root.c != Color.Black) {
			System.out.println(" root has color " + Color.Red);
			return -1;
		}
		
		int ctr = 0;
		Node its = firstEntry();
		while (its != null) {
			ctr++;
			Node child = its.left;
			if (child != null && child.parent != its) {
				System.out.println("Bad parent!");
			}
			child = its.right;
			if (child != null && child.parent != its) {
				System.out.println("Bad parent!");
			}
			
			Node nxt = successor(its);
			if (nxt == null) {
				break;
			}
			if (its.key.compareTo(nxt.key) >= 0) {
				System.out.println("Key order is wrong");
				return -3;
			}
			its = nxt;
		}
		int depth = verifyRedBlack (this.root, 0);
		//System.out.println("Nodes="+ctr+" Depth="+depth);
		return depth;
	}
	
	private static void insertInt (RBTree<Integer,Integer> tree, int data) {
		System.out.println("Inserting "+data);
		tree.put(data,data);
		System.out.println(tree);
		assert(tree.get(data) == data);
		assert(tree.verify() > 0);
	}
	
	private static void removeInt (RBTree<Integer,Integer> tree, int data) {
		System.out.println("Removing "+data);
		tree.remove(data);
		System.out.println(tree);
		assert(tree.get(data) == null);
		assert(tree.verify() > 0);
	}
	
	public static void main (String[] args) {
		int[] data = {3,1,4,1,5,9,2,6,5,3,5,8,9,7};
		
		System.out.println("Starting the test for the RBTree");
		RBTree<Integer,Integer> tree = new RBTree<Integer,Integer>();
		
		for (int i = 0; i < data.length; ++i) {
			insertInt(tree,data[i]);
		}
		
		data[0] = 2;
		data[1] = 3;
		data[2] = 1;
		for (int i = 0; i < data.length; ++i) {
			removeInt(tree,data[i]);
		}
		
		System.out.println("Done.");
	}
	*/
}
