/*Copyright (C) 2009  Annette Bieniusa, Arie Middelkoop.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package vacation;

//import java.util.Random;

import java.util.Iterator;
import java.util.NoSuchElementException;

import dstm.*;

@Transactional
public class SingleLinkedList<V extends Comparable<V>> implements Iterable<V>{
	
	
	private class SLLIterator implements Iterator<V> {

		Node current;
		
		public SLLIterator (Node n) {
			current = n;
		}
		
		@Atomic
		@Override
		public boolean hasNext() {
			return current.next != null;
		}

		@Atomic
		@Override
		public V next() {
			if (current != null) {
				current = current.next;
				return current.value;
			} else {
				throw new NoSuchElementException();
			}
		}

		@Atomic
		@Override
		public void remove() {
			throw new UnsupportedOperationException();
		}

		
	}

	@Transactional
	private class Node {
		V value;
		Node next;

		public Node(V i, Node n) {
			value = i;
			next = n;
		}

		public Node() {
			value = null;
			next = null;
		}
	}

	// head is sentinel node
	private Node head;

	@Atomic
	public void init() {
		Node n = new Node();
		this.head = n;
		return ;
		
	}
	
	
	@Atomic
	private Node previous(V value) {
		Node prev = this.head;
		Node curr = prev.next;

		while (curr != null) {
			if (value.compareTo(curr.value) <= 0) {
				return prev;
			}
			prev = curr;
			curr = curr.next;
		}
		return prev;
	}

	/**
	 * Inserts value into the list.
	 * 
	 * @param i
	 * @return true on successful insertion, false if value was already in the
	 *         list
	 */
	@Atomic
	public boolean add(V i) {
		Node prev = previous(i);
		Node curr = prev.next;

		// value is already in the list
		if (curr != null && curr.value.compareTo(i) == 0) {
			return false;
		}

		Node n = new Node(i, curr);
		prev.next = n;
		return true;
	}

	/**
	 * Searches for value in the list.
	 * 
	 * @param value
	 * @return value if found, otherwise null
	 */
	@Atomic
	public V find(V value) {
		Node prev = previous(value);
		Node n = prev.next;

		if (n == null || n.value.compareTo(value) != 0) {
			return null;
		} else {
			return n.value;
		}
	}

	/**
	 * Removes element of the list.
	 * 
	 * @param value
	 * @return true is element was removed, false if element has not been in the
	 *         list
	 */
	@Atomic
	public boolean remove (V value) {
		Node prev = previous(value);
		Node curr = prev.next;

		if (curr != null && curr.value.compareTo(value) == 0) {
			prev.next = curr.next;
			return true;
		} else {
			return false;
		}

	}

	/**
	 * Tests if list is empty.
	 * 
	 * @return true if list is empty, otherwise false
	 */
	@Atomic
	public boolean isEmpty() {
		return this.head == null;
	}

	/**
	 * Returns string representation of the list.
	 */
	@Atomic
	public String toString() {
		String s = "(";
		Node n = this.head.next;
		// empty list
		if (n == null) {
			s += ")";
		} else {
			s += n.value;
			while (n.next != null) {
				n = n.next;
				s += "," + n.value;
			}
			s += ")";
		}
		return s;
	}
	
	@Atomic
	public Iterator<V> iterator() {
		return new SLLIterator(head);
	}

	/*
	// Testing

	static int INSERTED_ELEMENTS = 5;
	static int UPPER_BOUND = 100000;
	static int NUM_THREADS = 1;
	static Random randomGen = new Random();

	public static class Tester extends Thread {
		final SingleLinkedList<Integer> l;

		public Tester(SingleLinkedList<Integer> l) {
			this.l = l;
		}

		public void run() {
			for (int i = 0; i < INSERTED_ELEMENTS; i++) {
				final int new_value = randomGen.nextInt(UPPER_BOUND) + 1;
				doInsert(new_value);
			}
		}

		boolean b(int x) {
			return false;
		}

		@Atomic
		private void doInsert(int new_value) {
			l.insert(new_value);
		}
	}

	public static void main(String[] args) {

		SingleLinkedList<Integer> list = new SingleLinkedList<Integer>();

		Tester threads[] = new Tester[NUM_THREADS];

		System.out.println("Initializing transactions");
		for (int i = 0; i < NUM_THREADS; ++i) {
			threads[i] = new Tester(list);
		}
		System.out.println("Starting transactions");
		for (int i = 0; i < NUM_THREADS; ++i) {
			threads[i].start();
		}

		try {
			for (int i = 0; i < NUM_THREADS; ++i) {
				threads[i].join();
			}
		} catch (InterruptedException e) {
			System.out.println("Urgh");
		}

		System.out.println(list.toString());
	}
	*/
}
